functor TableMIS(Table : TABLE where type Key.t = int) : MIS =
struct
   structure Set = Table.Set
   type vertex = Table.key
   type vertices = Table.set
   type graph = vertices Table.table
   structure Rand = Random210

   (* Returns a random integer given a random source rand and a vertex v 
      For the same source and vertex it will always return the same integer *)
   fun hashVertex (rand : Rand.rand) (v : vertex) 
      = Rand.hashInt rand v

   (* The neighbors of a single vertex *)
   fun N(G : graph, v : vertex) = 
     case (Table.find G v) of
       NONE => Set.empty
     | SOME(nbr) => nbr

   (* The neighbors of a set of vertices *)
   fun N'(G : graph, V : vertices) : vertices = 
       Table.reduce Set.union Set.empty (Table.extract(G,V))

   (* From a graph G = (V,E) extract a set of vertices V' subset V
      and the edges between them *)
   fun subGraph(G : graph, V' : vertices) : graph =
      Table.map (fn nbr => Set.intersection(nbr,V')) 
                (Table.extract(G,V'))

   (* The vertices of a graph *)
   fun Vtx(G : graph) : vertices = Table.domain G

   (* Returns a maximal independent set of the graph G *)
   fun MIS(G : graph) : vertices =
   let
     (* Given a random function R that maps vertices to integers,
        a graph G, and a vertex v, returns true if R(v) is greater
        than R(u) for all neighbors u of v*)
     fun isLocalMax (R : vertex -> int , G : graph) (v : vertex) =
       let
          val nbrVals = Table.tabulate R (N(G,v))
       in
          R(v) > Table.reduce Int.max 0 nbrVals
       end

     (* Returns a maximal independent set of the graph G based on
        a random source rand *)
     fun MIS'(G : graph, rand : Rand.rand, n) : vertices =
       if (Table.size G = 0) then Set.empty
       else
	 let
           val V = Vtx(G)
           val R = hashVertex rand
           
           (* Select local maximum values to add to MIS *)
           val inMIS = Set.filter (isLocalMax(R,G)) V

           (* Remove inMIS and their neighbors from V *)
           val V' = Set.difference(V, Set.union(inMIS, N'(G, inMIS)))

           (* Take subgraph based on just the remaining vertices V' *)
           val G' = subGraph(G, V')
         in
           Set.union(MIS'(G', Rand.next rand, n-1), inMIS)
         end

   in
     MIS'(G, Rand.fromInt 0, 10)
   end

end
